# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2017, 2018. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# ******************************************************************

import numpy as np
import sys
import math

from _snap_ml.utils import _is_mpi_enabled
_mpi_enabled = _is_mpi_enabled()

if sys.version_info[0] < 3:
    import libsnapmllocal2 as libsnapml
else:
    import _snap_ml.libsnapmllocal3 as libsnapml

class RandomForestClassifier:

    """
    Random Forest Classifier

    This class implements a random forest classifier using the IBM Snap ML library.
    It can be used for binary classification problems. It handles both dense and 
    sparse matrix inputs. Use csr, csc or ndarray matrix format for training and csr or 
    ndarray format for prediction. 

    Parameters
    ----------
    n_estimators : integer, optional, default : 10
        This parameter defines the number of trees in forest.
        
    criterion : string, optional, default : "gini"
        This function measures the quality of a split. The currently supported criterion is "gini".
    
    max_depth : integer or None, optional, default : None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than min_samples_leaf samples.

    min_samples_leaf : int or float, optional, default : 1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  
        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum number of samples for each node.
    
    max_features : int, float, string or None, optional, default : 'auto'
        The number of features to consider when looking for the best split:
            - If int, then consider `max_features` features at each split.
            - If float, then `max_features` is a fraction and
              `int(max_features * n_features)` features are considered at each
              split.
            - If "auto", then `max_features=sqrt(n_features)`.
            - If "sqrt", then `max_features=sqrt(n_features)`.
            - If "log2", then `max_features=log2(n_features)`.
            - If None, then `max_features=n_features`.
    
    bootstrap : boolean, optional, default : True
        This parameter determines whether bootstrap samples are used when building trees.

    n_jobs : integer or None, optional, default : None
        The number of jobs to run in parallel the fit function. None = 1 process.

    random_state : integer, or None, optional, default : None
        If integer, random_state is the seed used by the random number generator.
        If None, the random number generator is the RandomState instance used by `np.random`.    
    
    verbose : boolean, default : False
        If True, it prints debugging information while training. 
        Warning: this will increase the training time. For performance evaluation, use verbose=False.
    
    use_gpu : boolean, default : False
        Flag that indicates the hardware platform used for training. If True, the training 
        is performed using the GPU. If False, the training is performed using the CPU. Currently
        only CPU training is supported. 


    Attributes
    ----------

    """

    def __init__(self,  n_estimators=10,
                        criterion='gini', 
                        max_depth=None,
                        #min_samples_split=2,
                        min_samples_leaf=1,
                        max_features='auto',
                        #max_leaf_nodes=None,
                        bootstrap=True,
                        n_jobs=None,
                        random_state=None,
                        verbose=False, 
                        #class_weight=None, 
                        use_gpu=False):

        self.n_estimators_ = n_estimators
        if n_estimators <= 0:
            raise ValueError("Parameter n_estimators should be an integer >= 1.")
        
        self.criterion_ = criterion
        if (not criterion == 'gini'): #and (not criterion == 'entropy'):
            raise ValueError("Parameter criterion: invalid value. Valid values is 'gini'.")	

        self.max_depth_ = max_depth
        if (not max_depth is None) and (not isinstance(max_depth, (int))):
            raise ValueError("Parameter max_depth: invalid value or type. Valid values are 'None' or any integer > 0.")
        elif ((max_depth is not None) and max_depth <= 0):
            raise ValueError("Parameter max_depth: invalid value. Valid values are 'None' or any integer > 0.")
        
        #self.min_samples_split_ = min_samples_split
        #if (not isinstance(min_samples_split, (int))) and (not isinstance(min_samples_split, (float))):
        #    raise TypeError("Parameter min_samples_split: invalid type. Valid type is int (any integer >= 2) or non-negative float in (0,1].")
        #elif (min_samples_split <= 0):
        #    raise ValueError("Parameter min_samples_split: invalid value. Valid values are any integer >= 2 or non-negative float in (0,1].")
        #elif (min_samples_split > 1) and (isinstance(min_samples_split, (float))):
        #    raise ValueError("Parameter min_samples_split: invalid value. Valid values are any integer >= 2 or non-negative float in (0,1].")

        self.min_samples_leaf_ = min_samples_leaf
        if (not isinstance(min_samples_leaf, (int))) and (not isinstance(min_samples_leaf, (float))):
            raise TypeError("Parameter min_samples_leaf: invalid type. Valid type is int (any integer >= 1) or non-negative float in (0,1].")
        if (isinstance(min_samples_leaf, (int))) and (min_samples_leaf < 1):
            raise TypeError("Parameter min_samples_leaf: invalid value. Valid type is int (any integer >= 1) or non-negative float in (0,1].")
        if (isinstance(min_samples_leaf, (float))) and ((min_samples_leaf <= 0) or (min_samples_leaf > 1)):
            raise TypeError("Parameter min_samples_leaf: invalid value. Valid type is int (any integer >= 1) or non-negative float in (0,1].")

        self.max_features_ = max_features

        if (not max_features in [None, 'log2', 'sqrt', 'auto']):
            if (not isinstance(max_features, (int, float))):
                raise TypeError("Parameter max_features: invalid type. Valid type is int or float or string (None, log2, sqrt, auto).")
            else:
                if (isinstance(max_features, (int))) and (max_features < 1):
                    raise ValueError("Parameter max_features: invalid value. Valid value is any integer >= 1 or float in (0,1].")
                if (isinstance(max_features, (float))) and ((max_features > 1) or (max_features <= 0)):
                    raise ValueError("Parameter max_features: invalid value. Valid value is any integer >= 1 or float in (0,1].")
    
        #self.max_leaf_nodes_ = max_leaf_nodes 
        #if (not max_leaf_nodes is None) and (not isinstance(max_leaf_nodes, (int))):
        #    raise ValueError("Parameter max_leaf_nodes: invalid value or type. Valid values are 'None' or any integer >= 1.")
        #if (isinstance(max_leaf_nodes, (int))) and (max_leaf_nodes < 1):
        #    raise ValueError("Parameter max_leaf_nodes: invalid value. Valid values are 'None' or any integer >=1.")

        self.bootstrap_ = bootstrap
        if not isinstance(bootstrap, (bool)):
            raise TypeError("Parameter bootstrap: parameter type not valid. Should be a bool {True, False}.")

        self.n_jobs_ = n_jobs
        if (not n_jobs is None) and (not isinstance(n_jobs, (int))):
            raise ValueError("Parameter n_jobs: invalid value or type. Valid values are 'None' or any integer >= 1.")
            # TODO: scikit-learn -1 not yet covered
        
        self.random_state_ = random_state
        if (not isinstance(random_state, (int))) and (not random_state is None):
            raise ValueType("Parameter random_state: invalid value or type. Valid type is int or None.")

        self.verbose_ = verbose
        if not isinstance(verbose, (bool)):
            raise TypeError("Parameter verbose: invalid type. Valid type is bool {True, False}.")

        #self.class_weight_ = class_weight
        #if (not class_weight == 'balanced') and (not class_weight == None):
        #    raise ValueError("Parameter class_weight: invalid parameter value. Valid values are {'balanced', 'None'}.")

        self.use_gpu_ = use_gpu
        if not isinstance(use_gpu, (bool)):
            raise TypeError("Parameter use_gpu: invalid parameter type. Valid type is bool {False}.")
        elif use_gpu == True:
            raise ValueError("Parameter use_gpu: invalid parameter value. Valid value is {False}.")

        if _mpi_enabled:
            raise ValueError("MPI (distributed) implementation of decision tree classifiers not yet supported.")   
        
    def get_params(self):

        """
        Get the values of the model parameters.

        Returns
        -------
        params : dict
        """

        params = {}
        params['n_estimators'] = self.n_estimators_
        params['criterion'] = self.criterion_
        params['max_depth'] = self.max_depth_
        #params['min_samples_split'] = self.min_samples_split_
        params['min_samples_leaf'] = self.min_samples_leaf_
        params['max_features'] = self.max_features_
        #params['max_leaf_nodes'] = self.max_leaf_nodes_
        params['bootstrap'] = self.bootstrap_
        params['n_jobs'] = self.n_jobs_
        params['random_state'] = self.random_state_
        params['verbose'] = self.verbose_
        #params['class_weight'] = self.class_weight_
        params['use_gpu'] = self.use_gpu_
                
        return params

    def fit(self, X_train, y_train):

        """
        Fit the model according to the given train data.

        Parameters
        ----------
        X_train : sparse matrix (csr_matrix, csc_matrix) or dense matrix (ndarray)
            Train dataset

        y_train : array-like, shape = (n_samples,)
            The target vector corresponding to X_train. 

        Returns
        -------
        self : object
        """
        
                
        # Check if the gini criterion is used for evaluating the quality of a split
        if (self.criterion_ == 'gini'):
            gini_split_metric = True
        else:
            gini_split_metric = False

        # Check if the maximum depth is limited by the user, otherwise set value to 0 (no limit)
        if (self.max_depth_ is None):
            max_depth = 0 
        else:
            max_depth = self.max_depth_        

        # Check if the n_cpu parameter was set
        if (self.n_jobs_ is None):
            n_cpu = 1 
        else:
            n_cpu = self.n_jobs_
        
        # Random state
        if (self.random_state_ == None):
            # Not sure if this should be the random state
            random_state = np.random.get_state()[1][0]
        else:
            random_state = self.random_state_
            
        # Check if the balanced parameter was set
        #if (self.class_weight_ == 'balanced'):
        #    balanced = True
        #else:
        #    balanced = False    
            
        # Check value of max_leaf_nodes
        #if (self.max_leaf_nodes_ is None):
        #    # No limitation on the number of leaf_nodes
        #    max_leaf_nodes = 0    
        #else:
        #    max_leaf_nodes = self.max_leaf_nodes_     
              
        num_ft = 0
        num_nz = 0
        indptr = np.array([])
        indices = np.array([])
        data = np.array([])
        labs = np.array([])
        self.num_train_unique_labs = 0
        
        num_ex = len(y_train)

        # Check train dataset format
        if (X_train.shape[0] != num_ex):
            raise ValueError("X_train.shape[0] must be equal to number of examples.")

        if (type(X_train).__name__ == "csr_matrix"):
            num_ft = X_train.shape[1]
        elif (type(X_train).__name__ == "csc_matrix"):
            X_train = X_train.tocsr()
            num_ft = X_train.shape[1]
        elif (type(X_train).__name__ == "ndarray"):
            if (X_train.shape[0] == X_train.shape[1]):
                print("[Warning] Each row in the input train dataset is assumed to be a training sample.")
                num_ft = X_train.shape[0]
            elif (X_train.shape[0] == num_ex):
                num_ft = X_train.shape[1]
            else:
                X_train = X_train.transpose().copy()
                num_ft = X_train.shape[1]

        if (type(X_train).__name__ == "csr_matrix" or type(X_train).__name__ == "csc_matrix"):
            num_nz    = X_train.nnz
            indptr    = X_train.indptr.astype(np.uint64)
            data      = X_train.data.astype(np.float32)
            indices   = X_train.indices.astype(np.uint32)
        elif (type(X_train).__name__ == "ndarray"):
            num_nz    = num_ex * num_ft
            data      = X_train.astype(np.float32)
            indptr    = np.array([])
            indices   = np.array([])
        else:
            raise TypeError("X_train should be in csr_matrix, csc_matrix or ndarray format.")
    
        # Extract set of unique labels from y_train
        self.uniq_labs_ = np.unique(y_train)

        # Check the labels of y_train and train model(s)
        if (len(self.uniq_labs_) == 1):
            raise ValueError("There must be at least two unique label values in the train dataset.")
        elif (len(self.uniq_labs_) == 2):
            labs = y_train.astype(np.float32)
        else:
            raise ValueError("Multiclass classification not supported for decision tree classifiers.")

        # Check the value of min_samples_split
        #if (isinstance(self.min_samples_split_,(int))):
        #    if (self.min_samples_split_ > num_ex):
        #        raise ValueError("Parameter min_samples_split: invalid value. The value of min_samples_split is larger than the number of examples in the train dataset.")
        #    else:
        #        min_samples_split = self.min_samples_split_
        #else:
        #    min_samples_split = np.ceil(self.min_samples_split_ * num_ex)  
            
        # Check the value of min_samples_leaf
        if (isinstance(self.min_samples_leaf_,(int))):
            if (self.min_samples_leaf_ > num_ex):
                raise ValueError("Parameter min_samples_leaf: invalid value. The value of min_samples_leaf is larger than the number of examples in the train dataset.")
            else:
                min_samples_leaf = self.min_samples_leaf_
        else:
            min_samples_leaf = np.ceil(self.min_samples_leaf_ * num_ex)   
        
        # Check the value of max_features
        if (self.max_features_ is None):
            max_features = 0    # all features
        elif (self.max_features_ in ['auto', 'sqrt']):
            max_features = int(np.sqrt(num_ft)) 
        elif (self.max_features_ == 'log2'):
            import math
            max_features = int(math.log2(num_ft))
        elif (isinstance(self.max_features_, (int))):
            max_features = self.max_features_
        elif (isinstance(self.max_features_, (float))):
            max_features = int(self.max_features_*num_ft)
        else:
            raise ValueError("Parameter max_features: unable to parse.")

        max_features = int(max_features)
        if (max_features > num_ft):
            raise ValueError("Parameter max_features: invalid value. The value of max_features is larger than the number o features in the train dataset.")

        self.ptr_model_ = libsnapml.rfc_fit(self.n_estimators_,
                                    gini_split_metric, 
                                    max_depth,
                                    #min_samples_split,
                                    min_samples_leaf,
                                    max_features,
                                    #max_leaf_nodes,
                                    self.bootstrap_,                                    
                                    n_cpu,
                                    random_state,
                                    self.verbose_, 
                                    #balanced,                                    
                                    self.use_gpu_,
                                    num_ex, 
                                    num_ft,
                                    num_nz, 
                                    indptr, 
                                    indices, 
                                    data, 
                                    labs)

                                    
        return None

    # The test dataset should be in csr_matrix format
    def predict_proba(self, X, num_threads=0):

        """
        Probability estimates

        The returned probability estimates for the two classes.
        Only for binary classification. 

        Parameters
        ----------
        X : sparse matrix (csr_matrix) or dense matrix (ndarray)
            Dataset used for predicting probability estimates.

        num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.

        Returns
        -------
        None
        """
        
        raise ValueError("predict_proba is currently not supported for decision tree classifiers. Use predict to compute the class estimates.")
            
        return None

    def predict(self, X, num_threads=0):

        """
        Class predictions

        The returned class estimates.

        Parameters
        ----------
        X : sparse matrix (csr_matrix) or dense matrix (ndarray)
            Dataset used for predicting class estimates.

        num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.

        Returns
        -------
        proba: array-like, shape = (n_samples,)
            Returns the predicted class of the sample.
        """

        num_ex = 0
        num_ft = 0
        indptr = np.array([])
        indices =  np.array([])
        data = np.array([])        
        num_train_unique_labs = 0

        num_ex = X.shape[0]
        num_ft = X.shape[1]
        num_train_unique_labs = len(self.uniq_labs_)

        # libglm returns -1 and 1 as predicted classes
        # if the expected classes are 0 and 1, then we need to convert the predictions to 0 and 1
        # labels_flag keeps track of whether such a conversion is required
        labels_flag = False
        
        if (len(self.uniq_labs_) == 2) :
            if ((self.uniq_labs_[0] == 0) and (self.uniq_labs_[1] == 1)) or ((self.uniq_labs_[0] == 1) and (self.uniq_labs_[1] == 0)):
                labels_flag = True
        

        if (type(X).__name__ == "csc_matrix"):
            X = X.tocsr()
            indptr    = X.indptr.astype(np.uint64)
            data      = X.data.astype(np.float32)
            indices   = X.indices.astype(np.uint32)
        elif (type(X).__name__ == "csr_matrix"):
            indptr    = X.indptr.astype(np.uint64)
            data      = X.data.astype(np.float32)
            indices   = X.indices.astype(np.uint32)
        elif (type(X).__name__ == "ndarray"):
            data      = X.astype(np.float32)
            indptr    = np.array([])
            indices   = np.array([])
        else:
            raise ValueError("X should be in csc_matrix, csr_matrix or ndarray format.")

        pred = []

        # Generate predictions
        pred = libsnapml.rfc_predict(num_ex,
                                    num_ft,
                                    num_threads,
                                    indptr,
                                    indices,
                                    data,
                                    self.ptr_model_)

        pred = np.array(pred)

        # convert labels from -1 1 to 0 1
        if labels_flag:
            pred = (pred + 1)/2

        return pred

    # The test dataset should be in csr_matrix format
    def predict_log_proba(self, X, num_threads=0):

        """
        Log of probability estimates

        The returned log-probability estimates for the two classes.
        Only for binary classification.

        Parameters
        ----------
        X : sparse matrix (csr_matrix) or dense matrix (ndarray)
            Dataset used for predicting log-probability estimates.

        num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.

        Returns
        -------
        None
        """

        raise ValueError("predict_log_proba is currently not supported for decision tree classifiers. Use predict to compute the class estimates.")

        return None

